#! /usr/bin/env python3
"""
# snaphu.py is part of GMTSAR. 
# This Python script is migrated from snaphu.csh by Dunyu Liu on 20231109.
# snaphu.csh was originally written by X on X. 

# Purpose: to unwrap the phase.
"""

import sys, os, re, configparser
import subprocess, glob, shutil
from gmtsar_lib import * 

def snaphu():
    
    def Error_Message():
        print( "snaphu.py - unwrap the phase.")
        print( " if interp flag is invoked, unwrap the phase with nearest neighbor interpolating low coherence and blank pixels.")
        print( "Usage: snaphu.py correlation_threshold maximum_discontinuity interp [<rng0>/<rngf>/<azi0>/<azif>]")  
        print( " ")
        print( " correlation is reset to zero when < threshold ")
        print( " maximum_discontinuity enables phase jumps for earthquake ruptures, etc. ")
        print( " set maximum_discontinuity = 0 for continuous phase such as interseismic ")
        print( " interp=1, then calling nearest_grid to interpolate. ")
        print( " ")
        print( "Example: snaphu.py .12 40 1 1000/3000/24000/27000 ")
        print( "Reference: ")
        print( "Chen C. W. and H. A. Zebker, Network approaches to two-dimensional phase unwrapping: intractability and two new algorithms, Journal of the Optical Society of America A, vol. 17, pp. 401-414 (2000).")
        print( "Agram, P. S., & Zebker, H. A. (2009). Sparse two-dimensional phase unwrapping using regular-grid methods. IEEE Geoscience and Remote Sensing Letters, 6(2), 327-331.")
        
    print('SNAPHU - START ... ...')
    n = len(sys.argv)
    print('SNAPHU: input arguments are ', sys.argv)
    
    if n==1:
        print('SNAPHU: snaphu help information ... ...')
        Error_Message()
        sys.exit()
        
    if n<4:
        print('FILTER: Wrong # of input arguments; # should be larger than 3 ... ...')
        Error_Message()
    
    interp = int(sys.argv[3])
    if interp == 1:
        print('SNAPHU: interp is activated; unwrap the phase with nearest neighbor interpolating low coherence and blank pixels.')
    else: 
        print('SNAPHU: interp is NOT activated; unwrap the phase.')
        
    V = '-V'
    
    print('SNAPHU: prepare the files adding the correlation mask ... ...')
    if n==5:
        run('gmt grdcut mask.grd -R'+sys.argv[4]+' -Gmask_patch.grd')
        run('gmt grdcut corr.grd -R'+sys.argv[4]+' -Gcorr_patch.grd')
        run('gmt grdcut phasefilt.grd -R'+sys.argv[4]+' -Gphase_patch.grd')
    else:
        file_shuttle('mask.grd', 'mask_patch.grd', 'link')
        file_shuttle('corr.grd', 'corr_patch.grd', 'link')
        file_shuttle('phasefilt.grd', 'phase_patch.grd', 'link')
    
    print(' ')
    print('SNAPHU: ceate landmask ... ...')
    
    if check_file_report('landmask_ra.grd')==True:
        if n==5:
            par_tmp = subprocess.run(["gmt","grdinfo","-I","phase_patch.grd"],stdout=subprocess.PIPE).stdout.decode('utf-8').strip()
            print('SNAPHU: par_tmp is ', par_tmp)
            run('gmt grdsample landmask_ra.grd -R'+sys.argv[4]+' '+par_tmp+' -Glandmask_ra_patch.grd')
        else:
            if interp==0:
                run('gmt grdsample landmask_ra.grd -Rphase_patch.grd -Glandmask_ra_patch.grd')
            elif interp==1:
                par = catch_output_cmd(["gmt","grdinfo","-I","phase_patch.grd"],False,-999,-100000)
                run('gmt grdsample landmask_ra.grd '+par+' -Glandmask_ra_patch.grd')
        print(' ')
        run('gmt grdmath phase_patch.grd landmask_ra_patch.grd MUL = phase_patch.grd '+V)

    print(' ')
    print('SNAPHU: user defined mask ... ...')
    
    if check_file_report('mask_def.grd')==True:
        if n==5:
            run('gmt grdcut mask_def.grd -R'+sys.argv[4]+' -Gmask_def_patch.grd') 
        else:
            file_shuttle('mask_def.grd','mask_def_patch.grd','cp')
    
        print(' ')
        run('gmt grdmath corr_patch.grd mask_def_patch.grd MUL = corr_patch.grd '+V)

    
    run('gmt grdmath corr_patch.grd '+sys.argv[1]+' GE 0 NAN mask_patch.grd MUL = mask2_patch.grd')
    run('gmt grdmath corr_patch.grd 0. XOR 1. MIN  = corr_patch.grd')
    run('gmt grdmath mask2_patch.grd corr_patch.grd MUL = corr_tmp.grd') 
    
    if interp==0:
        run('gmt grd2xyz phase_patch.grd -ZTLf -do0 > phase.in')
    elif interp==1:
        run('gmt grdmath mask2_patch.grd phase_patch.grd MUL = phase_tmp.grd')
        run('nearest_grid phase_tmp.grd tmp.grd 300')
        file_shuttle('tmp.grd', 'phase_tmp.grd', 'mv')
        run('gmt grd2xyz phase_tmp.grd -ZTLf -do0 > phase.in')
    
    run('gmt grd2xyz corr_tmp.grd -ZTLf  -do0 > corr.in')
    
    print(' ')
    print('SNAPHU: run snaphu ... ...')
    
    sharedir = '/usr/local/GMTSAR/share/gmtsar'
    print(' ')
    print('SNAPHU: unwrapping phase with snaphu - higher threshold for faster unwrapping ... ...')
    
    par_tmp = catch_output_cmd(["gmt","grdinfo","-C","phase_patch.grd"], True, 10, -100000)
    #par_tmp = subprocess.run(["gmt","grdinfo","-C","phase_patch.grd"], stdout=subprocess.PIPE).stdout.decode('utf-8').strip().split()[9]
    print('SNAPHU: output from gmt grdinfo -C phase_patch.grd | cut -f 10 is ', par_tmp)
    
    if float(sys.argv[2]) == 0:    
        run('snaphu phase.in '+par_tmp+' -f '+sharedir+'/snaphu/config/snaphu.conf.brief -c corr.in -o unwrap.out -v -s -g conncomp.out')
    else:
        print('SNAPHU: replacing the line containing DEFOMAX_CYCLE to DEFOMAX_CYCLE $2 from snaphu.conf.brief... ...')
        file_shuttle(sharedir+'/snaphu/config/snaphu.conf.brief','snaphu.conf.brief','cp')
        replace_strings('snaphu.conf.brief','DEFOMAX_CYCLE','DEFOMAX_CYCLE '+sys.argv[2])
        run('snaphu phase.in '+par_tmp+' -f snaphu.conf.brief -c corr.in -o unwrap.out -v -d -g conncomp.out')
    
    print(' ')
    print('SNAPHU: convert to grd ... ...')
    
    par1 = catch_output_cmd(["gmt","grdinfo","-I-","phase_patch.grd"],False,0,-100000)
    par2 = catch_output_cmd(["gmt","grdinfo","-I", "phase_patch.grd"],False,0,-100000)
    print('SNAPHU: output from gmt grdinfo -I- phase_patch.grd is', par1)
    print('SNAPHU: output from gmt grdinfo -I phase_patch.grd is', par2)
    
    run('gmt xyz2grd unwrap.out -ZTLf -r '+par1+' '+par2+' -Gtmp.grd')
    print(' ')
    print('SNAPHU: generate connected component ... ...')
    run('gmt xyz2grd conncomp.out -ZTLu -r '+par1+' '+par2+' -Gconncomp.grd')
    run('gmt grdmath tmp.grd mask2_patch.grd MUL = tmp.grd')
    
    print(' ')
    print('SNAPHU: detrend the unwrapped if DEFOMAX = 0 for interseismic ... ...')
    file_shuttle('tmp.grd','unwrap.grd', 'mv')
    
    print(' ')
    print('SNAPHU: landmask ... ...')
    if check_file_report('landmask_ra.grd')==True:
        run('gmt grdmath unwrap.grd landmask_ra_patch.grd MUL = tmp.grd '+V)
        file_shuttle('tmp.grd','unwrap.grd', 'mv')
    
    print(' ')
    print('SNAPHU: user defined mask ... ...')
    if check_file_report('mask_def.grd')==True:
        run('gmt grdmath unwrap.grd mask_def_patch.grd MUL = tmp.grd '+V)
        file_shuttle('tmp.grd','unwrap.grd', 'mv')
    
    print(' ')
    print('SNAPHU: plot the unwrapped phase ... ...')
    
    run('gmt grdgradient unwrap.grd -Nt.9 -A0. -Gunwrap_grad.grd')
    tmp = catch_output_cmd(["gmt","grdinfo","-C","-L2","unwrap.grd"],True,-999,-100000)
    print('SNAPHU: output from cmd gmt grdinfo -C -L2 unwrap.grd is', tmp, 'which should be a list')
    limitU = float(tmp[11])+float(tmp[12])*2.
    limitU = round(limitU,1)
    limitL = float(tmp[11])-float(tmp[12])*2.
    limitL = round(limitL,1)
    std    = round(float(tmp[12]),1)
    run('gmt makecpt -Cseis -I -Z -T'+'''"'''+str(limitL)+'''"/"'''+str(limitU)+'''"/1 -D > unwrap.cpt''')
    
    if interp==1:
        tmp1 = catch_output_cmd(["gmt","grdinfo","unwrap.grd","-C"],True,-999,-100000)
        print('SNAPHU: output from cmd gmt grdinfo unwrap.grd -C is', tmp1, 'which should be a list')
        boundR = (float(tmp1[2])-float(tmp1[1]))/4
        boundA = (float(tmp1[4])-float(tmp1[3]))/4
        
    run('gmt grdimage unwrap.grd -Iunwrap_grad.grd -Cunwrap.cpt -JX6.5i -Bxaf+lRange -Byaf+lAzimuth -BWSen -X1.3i -Y3i -P -K > unwrap.ps')
    run('''gmt psscale -Runwrap.grd -J -DJTC+w5/0.2+h+e -Cunwrap.cpt -Bxaf+l"Unwrapped phase" -By+lrad -O >> unwrap.ps''')
    run('gmt psconvert -Tf -P -A -Z unwrap.ps')
    
    print(' ')
    print('SNAPHU: unwrapped phase map: unwrap.pdf ... ...')
    
    print(' ')
    print('SNAPHU: clean up ... ...')
    
    run('rm -f tmp.grd corr_tmp.grd unwrap.out tmp2.grd unwrap_grad.grd conncomp.out')
    run('rm -f phase.in corr.in') 
    
    if interp==1:
        run('rm -f phase_tmp.grd')
        file_shuttle('phase_patch.grd','phasefilt_interp.grd','mv')
        
    if n==5:
        file_shuttle('corr_patch.grd', 'corr_cut.grd', 'mv')
    run('rm -f mask_patch.grd mask3.grd mask3.out')
    run('rm -f corr_cut.grd corr_patch.grd')
    
    print("SNAPHU - END ... ...")

def _main_func(description):
    snaphu()

if __name__ == "__main__":
    _main_func(__doc__)




#! /usr/bin/env python3
import os, time
from pathListForTest import caseNameList, intfDirList, rawDir, \
    SLCDir, workAbsoluteDir, pythonCommandListPath

for caseName in caseNameList:
    startTime = time.time()
    os.chdir(workAbsoluteDir+caseName)
    os.system('pwd')
    os.system('cleanup all')
    os.system('cp -r '+pythonCommandListPath+'README_'+caseName+'.txt .')
    os.system('./README_'+caseName+'.txt > log.txt')
    elapsedTime = time.time()-startTime
    with open('timeSpentLog.txt', 'a') as f:
        f.write(caseName+' used '+str(elapsedTime)+' s \n')

os.chdir(workAbsoluteDir)
os.system('python checkTest.py')



#!/usr/bin/python

import subprocess
import os.path
import numpy as np
import sys

def get_trend(x,y,z):
    G=np.column_stack((np.ones(np.shape(x)),x,y))
    #p=(G'*G)\G'*z
    p=np.dot(np.dot(np.linalg.inv(np.dot(np.transpose(G),G)),np.transpose(G)),z)
    return p

if __name__ == '__main__':

    # load data from the unwrapped subset
    unwrap=sys.argv[1] # eg. unwrap.dat
    data=np.loadtxt(unwrap)
    x=data[:,0]
    y=data[:,1]
    z=data[:,2]

    # find the trend
    trend_params=get_trend(x,y,z)
    print('%.15f %.15f %.15f') %(trend_params[0], trend_params[1],trend_params[2])


    
## END ##



#! /usr/bin/env python3
import os, time
import numpy as np
import xarray as xr
from skimage import io
from skimage.metrics import structural_similarity as ssim
import matplotlib.pyplot as plt
from pathListForTest import caseNameList, intfDirList, rawDir, SLCDir

fileNameList = ['corr_ll.png','display_amp_ll.png','phasefilt_mask_ll.png',
        'corr_ll.grd', 'phasefilt.grd', 'filtcorr.grd']
refRoot  = 'csh.test'
testRoot = 'py.test'
fileDiffNumericThreshold = 1e-3
imageSimilarityIndexThreshold = 0.999

def parseCmdOutput(fn, searchStr):
    with open(fn,'r') as f:
        for line in f:
            if searchStr in line:
                val = line.split()
                keyIndex = val.index(searchStr)
                result = float(val[keyIndex+1])
    return result

def compare_nc_files(fn1,fn2,threshold=1e-3):
    isTheSame = 'SUCCESS '+fn1+' '+fn2
    f1 = xr.open_dataset(fn1)
    f2 = xr.open_dataset(fn2)
    metadata_equal = f1.identical(f2)
    data_equal = (f1==f2).all().items()
    
    # Compare variables
    for var in f1.variables:
        var1 = f1[var]
        var2 = f2[var]
        if var1.dims != var2.dims:
            isTheSame = 'FAIL var dim '+fn1+' '+fn2
        if not np.allclose(var1,var2,rtol=threshold, atol=threshold):
            isTheSame = 'FAIL var numbers '+fn1+' '+fn2
        
    if not metadata_equal:# and data_equal:
        isTheSame = 'FAIL metadata '+fn1+' '+fn2
    
    try:
        xr.testing.assert_allclose(f1,f2)
        print('SUCCESS by xarray.testing.assert_allclose')
        print('SUCCESS '+fn1 +' '+fn2)
    except AssertionError as e:
        print(e)
    print(isTheSame)
    
    return isTheSame

def compare_txt_files(fn1,fn2,threshold=1e-3):
    isTheSame = 'SUCCESS '+fn1+' '+fn2
    with open(fn1,'r') as f1, open(fn2,'r') as f2:
        result1 = f1.read().split()
        result2 = f2.read().split()
    if len(result1) != len(result2):
        isTheSame = 'FAIL '+fn1+' '+fn2
    
    for num1,num2 in zip(result1,result2):
        fnum1, fnum2 = float(num1),float(num2)
        if abs(fnum1-fnum2) > threshold:
            isTheSame = 'FAIL '+fn1+' '+fn2
    print(isTheSame)

def compare_files(fnNew, fnRef, fileName, fileType):
    isTheSame = 'SUCCESS: python and csh '+fileName+' are the same'
    notTheSame = 'FAIL: python and csh '+fileName+' are different'

    if fileType=='png':
        imageNew = io.imread(fnNew)
        imageRef = io.imread(fnRef)
        #assert imageNew.shape == imageRef.shape, 'Images must be the same shape.'
        try:
            ssim_index = ssim(imageNew,imageRef,multichannel=True)
            if ssim_index>imageSimilarityIndexThreshold:
                print(isTheSame+' '+f'SSIM: {ssim_index}')
            else:
                print(notTheSame+' '+f'SSIM: {ssim_index}')
        except:
            print(notTheSame+' no SSIM')

        if imageNew.shape != imageRef.shape:
            print(notTheSame+' image shapes do not match')
    elif fileType=='grd':
        os.system('gmt grdmath '+fnNew+' '+fnRef+' SUB = diff.grd')
        os.system('gmt grdinfo -L2 diff.grd > gmt.log.txt')
        mean = parseCmdOutput('gmt.log.txt', 'mean:')
        stdev = parseCmdOutput('gmt.log.txt', 'stdev:')
        rms = parseCmdOutput('gmt.log.txt', 'rms:')
        #print('Python and csh '+fileName+' "s difference are')
        #print('mean = ',mean, ' stddev = ', stdev, 'rms = ', rms)
        if (rms<fileDiffNumericThreshold and not('phase' in fnNew)) or \
                (rms<0.1 and ('phase' in fnNew)):
            print(isTheSame+'; diff.grd mean='+str(mean)+' stdev='+str(stdev)+' rms='+str(rms))
        else:
            print(notTheSame+' diff.grd mean='+str(mean)+' stdev='+str(stdev)+' rms='+str(rms))
        os.system('rm diff.grd gmt.log.txt')

def findErrorsInLogFiles(rootDir):
    errKeyWordList = ['error', 'Error', 'Traceback', 'ERROR']
    for root, dirs, files in os.walk(rootDir):
        for file in files:
            if file=='log.txt':
                with open(os.path.join(root,file),'r') as f:
                    contents = f.read()
                    if any(errKeyWord in contents for errKeyWord in errKeyWordList):
                        print('Error found in ', os.path.join(root,file))
                    else:
                        print('No Error found in ', os.path.join(root,file))


for caseName in caseNameList:
    print(' ')
    print('Comparing case ', caseName)
    for fileName in fileNameList:
        for path in intfDirList[caseName]:
            refPath  = refRoot+'/'+caseName+'/'+path+'/'+fileName
            testPath = testRoot+'/'+caseName+'/'+path+'/'+fileName
            if os.path.exists(refPath) and os.path.exists(testPath):
                if 'png' in fileName:
                    compare_files(testPath, refPath, fileName, 'png')
                elif 'grd' in fileName:
                    compare_files(testPath, refPath, fileName, 'grd')

    findErrorsInLogFiles(testRoot+'/'+caseName)




#! /usr/bin/env python3
"""
# gmtsar_lib.py is part of pyGMTSAR. 
# It hosts commonly used functions similar to CSH.
# Dunyu Liu, 20230202.

# check_file_report
# grep_value
# replace_strings
# file_shuttle
"""

########################################
    ## The following block loads configuration files from config.ini
    ## Currently keep it but may phase out in the future.
    ########################################
    
    # print('P2P 0: load config.ini ... ...')
    # config   = configparser.ConfigParser()
    # config.read('config.ini')
    
    # # stage, which stage to process, int.
    # stage    = int(config['params']['proc_stage'])
    # s_stage  = config['params']['skip_stage']
    # #s_stage= config
    # print(" ")
    # print('P2P 0: proc_stage = ', stage)
    # print('P2P 0: skip_stage = ', s_stage)

    # skip_1   = 0
    # skip_2   = 0
    # skip_3   = 0
    # skip_4   = 0
    # skip_5   = 0
    # skip_6   = 0
    
    # if s_stage != '':
        # print(" ")
        # print("P2P 0: skipping stages are ", s_stage, ' ... ...')
    
    # tmp = config['params']['skip_master'] # tmp is a string. Needs to convert it to numbers for manipulation later.
    # if tmp == '':
        # skip_master = 0
    # else:
        # skip_master = int(tmp)

    # if skip_master == 2:
        # skip_4 = 1
        # skip_5 = 1
        # skip_6 = 1
        # print("P2P 0: skipping stages 4,5, and 6 as skip_master is set to 2 ...")
    
    # print("P2P 0: skip_master = ", skip_master)
    
    # print('P2P 0: loading num_patches, near_range, earth_radius, fd (floats)')
    # num_patches   = config['params']['num_patches']
    # near_range    = config['params']['near_range']
    # earth_radius  = config['params']['earth_radius']
    # fd            = config['params']['fd1']
    
    # print('P2P 0: loading topo_phase, topo_interp_mode (ints)')
    # topo_phase    = int(config['params']['topo_phase'])
    # topo_interp_mode = int(config['params']['topo_interp_mode'])

    # if topo_interp_mode == '':
        # topo_interp_mode = 0

    # shift_topo    = int(config['params']['shift_topo'])
    # switch_master = int(config['params']['switch_master'])
    # filter        = config['params']['filter_wavelength']
    # compute_phase_gradient = config['params']['compute_phase_gradient']
    # iono          = config['params']['correct_iono']
    # if iono == '':
        # iono = 0
    
    # iono_filt_rng = config['params']['iono_filt_rng']
    # iono_filt_azi = config['params']['iono_filt_azi']
    # iono_dsamp    = config['params']['iono_dsamp']
    # iono_skip_est = config['params']['iono_skip_est']
    
    # # spec_div could be missing.
    # if config.has_option('params','spec_div') == True:
        # spec_div      = config['params']['spec_div']
    # else:
        # spec_div      = 0
    
    # # spec_mode could be missing.
    # if config.has_option('params','spec_mode') == True:
        # spec_mode     = config['params']['spec_mode']
    # else:
        # spec_mode     = ''
        
    # dec           = config['params']['dec_factor']
    # threshold_snaphu  = float(config['params']['threshold_snaphu'])
    # threshold_geocode = float(config['params']['threshold_geocode'])
    # #if config.has_option('params','region_cut') == True:
    # region_cut    = config['params']['region_cut']
    # #else:
    # print('P2P 0: region_cut =', region_cut)
    # print((region_cut==''))
    # print((region_cut==None))
    
    # mask_water    = config['params']['mask_water']
    # print('P2P 0: mask_water =', mask_water)
    
    # # switch_land could be missing 
    # if config.has_option('params','switch_land')  == True:
        # switch_land   = config['params']['switch_land']
    # else: 
        # switch_land   = ''
    # print('P2P 0: switch_land =', switch_land)    
    # defomax       = config['params']['defomax']
    
    # # range_dec could be missing
    # if config.has_option('params','range_dec')    == True:
        # range_dec     = config['params']['range_dec']
    # else:
        # range_dec     = ''
        
    # if config.has_option('params','range_dec')    == True:
        # azimuth_dec   = config['params']['azimuth_dec']
    # else:
        # azimuth_dec   = ''
    # if config.has_option('params', 'SLC_factor')  == True:
        # SLC_factor    = config['params']['SLC_factor']
    # else: 
        # SLC_factor    = ''
        
    # near_interp   = config['params']['near_interp']
    # master        = sys.argv[2]
    # aligned       = sys.argv[3]

    # print(config.items('params'))
    # print(' ')


#!/usr/bin/env python3

import numpy as np
import requests
import re
import os
import argparse
import datetime
from html.parser import HTMLParser

server = 'https://scihub.copernicus.eu/gnss/'

orbitMap = [('precise', 'AUX_POEORB'),
            ('restituted', 'AUX_RESORB')]

datefmt = "%Y%m%dT%H%M%S"
queryfmt = "%Y-%m-%d"
queryfmt2 = "%Y/%m/%d/"

#Generic credentials to query and download orbit files
credentials = ('gnssguest', 'gnssguest')


def cmdLineParse():
    '''
    Command line parser.
    '''

    parser = argparse.ArgumentParser(description='Fetch orbits corresponding to given SAFE package')
    parser.add_argument('-i', '--input', dest='input', type=str, required=True,
                        help='Path to SAFE package of interest')
    parser.add_argument('-o', '--output', dest='outdir', type=str, default='.',
                        help='Path to output directory')

    return parser.parse_args()


def FileToTimeStamp(safename):
    '''
    Return timestamp from SAFE name.
    '''
    safename = os.path.basename(safename)
    fields = safename.split('_')
    sstamp = []  # sstamp for getting SAFE file start time, not needed for orbit file timestamps

    try:
        tstamp = datetime.datetime.strptime(fields[-4], datefmt)
        sstamp = datetime.datetime.strptime(fields[-5], datefmt)
    except:
        p = re.compile(r'(?<=_)\d{8}')
        dt2 = p.search(safename).group()
        tstamp = datetime.datetime.strptime(dt2, '%Y%m%d')

    satName = fields[0]

    return tstamp, satName, sstamp


class MyHTMLParser(HTMLParser):

    def __init__(self,url):
        HTMLParser.__init__(self)
        self.fileList = []
        self._url = url
        
    def handle_starttag(self, tag, attrs):
        for name, val in attrs:
            if name == 'href':
                if val.startswith("https://scihub.copernicus.eu/gnss/odata") and val.endswith(")/"):
                    pass
                else:
                    downloadLink = val.strip()
                    downloadLink = downloadLink.split("/Products('Quicklook')")
                    downloadLink = downloadLink[0] + downloadLink[-1]
                    self._url = downloadLink
                
    def handle_data(self, data):
        if data.startswith("S1") and data.endswith(".EOF"):
            self.fileList.append((self._url, data.strip()))
            

def download_file(url, outdir='.', session=None):
    '''
    Download file to specified directory.
    '''

    if session is None:
        session = requests.session()

    path = outdir
    print('Downloading URL: ', url)
    request = session.get(url, stream=True, verify=True, auth=credentials)

    try:
        val = request.raise_for_status()
        success = True
    except:
        success = False

    if success:
        with open(path, 'wb') as f:
            for chunk in request.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
                    f.flush()

    return success


def fileToRange(fname):
    '''
    Derive datetime range from orbit file name.
    '''

    fields = os.path.basename(fname).split('_')
    start = datetime.datetime.strptime(fields[-2][1:16], datefmt)
    stop = datetime.datetime.strptime(fields[-1][:15], datefmt)
    mission = fields[0]

    return (start, stop, mission)


if __name__ == '__main__':
    '''
    Main driver.
    '''

    inps = cmdLineParse()

    fileTS, satName, fileTSStart = FileToTimeStamp(inps.input)
    print('Reference time: ', fileTS)
    print('Satellite name: ', satName)
    match = None
    session = requests.Session()

    for spec in orbitMap:
        oType = spec[0]
        delta = datetime.timedelta(days=1)
        timebef = (fileTS - delta).strftime(queryfmt)
        timeaft = (fileTS + delta).strftime(queryfmt)
        url = server + 'search?q=( beginPosition:[{0}T00:00:00.000Z TO {1}T23:59:59.999Z] AND endPosition:[{0}T00:00:00.000Z TO {1}T23:59:59.999Z] ) AND ( (platformname:Sentinel-1 AND filename:{2}_* AND producttype:{3}))&start=0&rows=100'.format(timebef,timeaft, satName,spec[1])
        
        success = False
        match = None
  
        try:
            r = session.get(url, verify=True, auth=credentials)
            r.raise_for_status()
            parser = MyHTMLParser(url)
            parser.feed(r.text)
            for resulturl, result in parser.fileList:
                tbef, taft, mission = fileToRange(os.path.basename(result))
                if (tbef <= fileTSStart) and (taft >= fileTS):
                    matchFileName = result
                    match = resulturl

            if match is not None:
                success = True
        except:
            pass

        if success:
            break

    if match is not None:
        output = os.path.join(inps.outdir, matchFileName)
        res = download_file(match, output, session)
        if res is False:
            print('Failed to download URL: ', match)
    else:
        print('Failed to find {1} orbits for tref {0}'.format(fileTS, satName))



#! /usr/bin/env python3
"""
# gmtsar_lib.py is part of pyGMTSAR. 
# It hosts commonly used functions similar to CSH.
# Dunyu Liu, 20230202.

# check_file_report
# grep_value
# replace_strings
# file_shuttle
"""

import sys, os, re, configparser
import subprocess, glob

def check_file_report(fn):
    # Check if a file exists.
    # If not, print error message.
    #
    exist = True
    if os.path.isfile(fn) == False:
        exist = False
        print(" no file " + fn)
        #sys.exit()
    return exist

def catch_output_cmd(cmd_list, choose_split=False, split_id=-999, digit_id=-100000):
    # catch_output_cmd takes in cmd_list and return the string
    tmp = subprocess.run(cmd_list, stdout=subprocess.PIPE).stdout.decode('utf-8').strip()
    
    if choose_split==True:
        
        if split_id==-999:
            out = tmp.split() # return a list
        else:
            out = tmp.split()[split_id-1] # return a value
            
            if digit_id!=-100000:
                out = tmp.split()[split_id-1][digit_id-1]
    else:
        out = tmp 
        # If choose_split==False, default return is a string. 
    return out

def intFloatOrString(val):
    if val.isdigit():
        return int(val)
    else:
        try:
            return float(val)
        except ValueError:
            return ""
            
def grep_value(fn, s, i): 
    # grep_value performs similar functions to unix grep. 
    # Given a file name - fn, and a character string - s, find the ith value.
    # The character should be unique in file fn.
    #
    with open(fn, 'r') as f:
        for line in f.readlines():
            if re.search(s, line):
                print(line)
                val = line.split()[i-1]     
    return intFloatOrString(val)

def replace_strings(fn, s0, s1):
    # replace_strings will replace str s0 in file fn0,
    #   with the string s1, and update fn0.
    with open(f"{fn}") as f:
        lines = f.readlines()

    updated_lines = []
    for line in lines:
        if s0 in line:
            line = f"{s1}\n"
        updated_lines.append(line)

    with open(f"{fn}", "w") as f:
        f.writelines(updated_lines)

def append_new_line(fn,s0):
    # append the string s0 as a new line at the end of file named fn.
    with open(fn,"a+") as f:
        f.seek(0)
        data = f.read(100)
        if len(data)>0:
            f.write("\n")
        f.write(s0)

def file_shuttle(fn0, fn1, opt):
    # copy/move fn0 and paste it to fn1.
    if opt == "cp":
        print("cp " + fn0 + " " + fn1)
        os.system("cp " + fn0 + " " + fn1)
    elif opt == "mv":
        print("mv " + fn0 + " " + fn1)
        os.system("mv " + fn0 + " " + fn1)
    elif opt == "link":
        print("ln -sf " + fn0 + " " + fn1)
        os.system("ln -sf " + fn0 + " " + fn1)

def delete(fn):
    # delete file named fn.
    os.system("rm -rf "+fn)
    
def assign_arg(arg, str):
    # arg is the list that contains arguments from a terminal input.
    # the function will search for string specified in 'str', and 
    # return the value next to it in arg.
    if str in arg:
       val = arg[arg.index(str)+1]
       return intFloatOrString(val)
    else:
       return 0

def run(cmd):
    # run and print the command specified in cmd.
    print(" ")
    print(cmd)
    os.system(cmd)

def renameMasterAlignedForS1tops(master0, aligned0):
    print('Renaming master and aligned for SAT==S1_TOPS')
    master = 'S1_'+master0[15:15+8]+'_'+master0[24:24+6]+'_F'+master0[6:7]
    aligned = 'S1_'+aligned0[15:15+8]+'_'+aligned0[24:24+6]+'_F'+aligned0[6:7]
    return master, aligned